# -- import packages: ---------------------------------------------------------
import ABCParse
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np


# -- set typing: --------------------------------------------------------------
from typing import List


# -- operational object class: ------------------------------------------------
class TemporalColorMapFactory(ABCParse.ABCParse):
    def __init__(self, cmap: matplotlib.colors.ListedColormap = cm.plasma_r):
        self.__parse__(locals())

    @property
    def _original_cmap(self):
        return np.array(self._cmap.colors)

    @property
    def cmap(self) -> List[List[float]]:
        _cmap = self._original_cmap[:: self._idx_slice][
            self._pad_left : -self._pad_right
        ]
        return matplotlib.colors.ListedColormap(_cmap, name="TemporalColorMap")

    def __call__(
        self,
        idx_slice: int = 6,
        pad_left: int = 1,
        pad_right: int = 1,
        plot: bool = False,
    ):
        self.__update__(locals())

        return self.cmap


# -- API-facing function: -----------------------------------------------------
def generate_temporal_cmap(
    idx_slice: int = 6,
    pad_left: int = 1,
    pad_right: int = 1,
    plot: bool = False,
    cmap: matplotlib.colors.ListedColormap = cm.plasma_r,
):
    time_cmap = TemporalColorMapFactory()
    return time_cmap(
        idx_slice=idx_slice, pad_left=pad_left, pad_right=pad_right, plot=plot
    )


# -- API-facing object: --------------------------------------------------------
temporal_colormap = generate_temporal_cmap()



# -- import packages: ---------------------------------------------------------
import pathlib

# -- import local dependencies: -----------------------------------------------
from ._screen_result import ScreenResult

# -- set type hints: ----------------------------------------------------------
from typing import Iterator

# -- operational object cls: --------------------------------------------------
class ScreenResults:
    """Fate perturbation screen results"""
    def __init__(self, csv_dir: str) -> None:
        self._csv_dir = pathlib.Path(csv_dir)
        self._paths = list(self._csv_dir.glob("*"))
        self._get_data()

    def _get_versions(self) -> None:
        self._versions = []
        for path in self._paths:
            v = path.name.split(".")[0]
            if not v in self._versions:
                self._versions.append(v)
        self._versions = sorted(self._versions)

    def _get_data(self) -> None:
        self._get_versions()
        for version in self._versions:
            if not hasattr(self, version):
                base_path = self._csv_dir.joinpath(version)
                screen_result = ScreenResult(base_path)
                setattr(self, version, screen_result)

    def __repr__(self) -> str:
        return f"ScreenResults(csv_dir={self._csv_dir})"

    def __str__(self) -> str:
        return f"ScreenResults(csv_dir={self._csv_dir})"

    def __len__(self) -> int:
        return len(self._versions)
    
    def __getitem__(self, version: int) -> ScreenResult:
        return getattr(self, f"version_{version}")
    
    def __iter__(self) -> Iterator[ScreenResult]:
        for version in self._versions:
            yield getattr(self, version)
    
    def __contains__(self, version: str) -> bool:
        return version in self._versions



from ._box_plot import BoxPlot
from ._styled_box_plot import StyledBoxPlot
from ._foreground_scatter import ForegroundScatter

from ._styled_box_plot import StyledBoxPlot, boxplot

__all__ = ['StyledBoxPlot', 'boxplot']


from ._compute_cytotrace_pseudotime import compute_CytoTRACE_pseudotime
from ._bin_pseudotime import bin_pseudotime
from ._subset_lineage_adata import subset_lineage_adata


__all__ = [
    "compute_CytoTRACE_pseudotime",
    "bin_pseudotime",
    "subset_lineage_adata",
]




# -- import packages: ---------------------------------------------------------
import anndata

# -- set typing: --------------------------------------------------------------
from typing import Union, Tuple
# CytoTRACEKernel = cellrank.kernels.CytoTRACEKernel


# -- API-facing function: -----------------------------------------------------
def compute_CytoTRACE_pseudotime(
    adata: anndata.AnnData, return_kernel: bool = False,
): # -> Union[anndata.AnnData, Tuple[anndata.AnnData, CytoTRACEKernel]]:
    """Use cellrank to compute CytoTRACE"""

    import scvelo
    import cellrank
    
    # -- preprocess with scvelo: ----------------------------------------------
    scvelo.pp.moments(adata)

    # -- compute CytoTRACE time with cellrank: --------------------------------
    adata.layers["Ms"] = adata.X
    ctk = cellrank.kernels.CytoTRACEKernel(adata)
    ctk.compute_cytotrace()

    if return_kernel:
        return adata, ctk
    return adata



# -- import packages: --
import ABCParse
import numpy as np
import pandas as pd


# -- operational cls: ------
class CrossEntropy(ABCParse.ABCParse):
    def __init__(self, epsilon: float = 1e-12):
        self.__parse__(locals())
        
    def _augment_cols(self, F_obs, F_hat):
        for col in F_obs:
            if not col in F_hat:
                F_hat[col] = 0
        return F_hat
        
    def _norm(self, df: pd.DataFrame):
        return df.div(df.sum(axis=1), axis = 0)
    
    def _filter_zero_prediction_rows(self, F_obs: pd.DataFrame, F_hat: pd.DataFrame):
        mask = (F_hat.sum(1) > 0).values
#         print(mask.sum())
        F_obs = F_obs.loc[mask]
        F_hat = F_hat.loc[mask]
        return F_obs, F_hat, mask
    
    def __call__(self, F_obs, F_hat):
        
        true = F_obs.copy()
        pred = F_hat.copy()
        
            
        pred = self._augment_cols(F_obs = true, F_hat = pred)
        pred = pred[true.columns.tolist()]
        
#         true, pred, mask = self._filter_zero_prediction_rows(F_obs = true, F_hat = pred)    
        
        true['undiff'] = 0
        pred['undiff'] = 1 - pred.sum(1)
        
        
        pred = pred.clip(lower=self._epsilon)
        true = self._norm(df=true)
        pred = self._norm(df=pred)
        
        
        ce = -np.sum(true.values * np.log(pred.values + self._epsilon), axis=1)       
#         n_cells = mask.sum()
        return np.abs(ce).mean()

# -- api-facing function: ----    
def cross_entropy(F_obs: pd.DataFrame, F_hat: pd.DataFrame, epsilon: float = 1e-12):
    """"""
    calc = CrossEntropy(epsilon=epsilon)
    return calc(F_obs=F_obs, F_hat=F_hat)



# -- import packages: ----------------------------------------------------------
import anndata
import logging
import pathlib

# -- import local dependencies: -----------------------------------------------
from ._annotate_adata_with_growth_weights import annotate_adata_with_growth_weights

# -- set type hints: ----------------------------------------------------------
from typing import Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- define function: ---------------------------------------------------------
def load_adata(
    h5ad_path: Union[pathlib.Path, str],
    weight_key: str = "W",
    time_key: str = "Time point",
    time_point: int = 2,
) -> None:

    logger.info(f"Loading adata from {h5ad_path}...")
    adata = anndata.read_h5ad(h5ad_path)

    # -- UPDATE TO TRAIN ONLY ON WELLS 0, 1: ----------------------------
    adata.obs["train"] = adata.obs["Well"].isin([0, 1])

    logger.info("Annotating adata with growth weights...")
    adata = annotate_adata_with_growth_weights(
        adata=adata,
        weight_key=weight_key,
        time_key=time_key,
        time_point=time_point,
    )
    logger.info(f"\nAnnData prepared:\n{adata}")
    return adata



# -- import packages: ---------------------------------------------------------
import ABCParse
import pandas as pd
import scdiffeq as sdq


# -- supporting class: --------------------------------------------------------
class PerEpochLoss(ABCParse.ABCParse):
    """Tabulator of per epoch loss"""

    def __init__(
        self,
        t_ignore="0.0",
        epoch_key: str = "epoch",
        metric: str = "sinkhorn",
        *args,
        **kwargs,
    ) -> None:
        """
        Args:
            t_ignore (str) ***Default**: "0.0"

            epoch_key (str) ***Default**: "epoch"

            metric (str) ***Default**: "sinkhorn"

        Returns:
            None
        """
        self.__parse__(locals())

    def _aggregate(self, df: pd.DataFrame):
        """Filter, sum, mean: apply, per epoch to grouped df"""
        return (
            df.filter(regex=self._stage)
            .filter(regex=self._metric)
            .drop(f"{self._metric}_{self._t_ignore}_{self._stage}", axis=1)
            .dropna()
            .sum(1)
            .mean()
        )

    @property
    def _EPOCH_GROUPED_METRICS(self):
        """ """
        if not hasattr(self, "_GROUPED_BY"):
            self._GROUPED_BY = self._metrics_df.groupby(self._epoch_key)
        return self._GROUPED_BY

    def forward(self):
        """ """
        return self._EPOCH_GROUPED_METRICS.apply(self._aggregate) # , include_groups=False)

    def __call__(
        self, metrics_df: pd.DataFrame, stage: str = "training", *args, **kwargs
    ) -> pd.Series:
        """
        Args:
            metrics_df (pd.DataFrame)

            stage (str)
        Returns:
            loss (pd.Series)
        """
        self.__update__(locals())

        return self.forward()


# -- operational class: -------------------------------------------------------
class SummarizedLoss(ABCParse.ABCParse):
    def __init__(
        self,
        version: sdq.io.Version,
        n_best: int = 10,
        training_key: str = "training",
        validation_key: str = "validation",
        t_ignore="0.0",
        epoch_key: str = "epoch",
        metric: str = "sinkhorn",
        *args,
        **kwargs
    ):
        self.__parse__(locals())

    @property
    def metrics_df(self) -> pd.DataFrame:
        if not hasattr(self, "_metrics_df"):
            self._metrics_df = self._version.metrics_df
        return self._metrics_df

    @property
    def _PER_EPOCH_LOSS_TABULATOR(self):
        if not hasattr(self, "_per_epoch_loss_tabulator"):
            self._per_epoch_loss_tabulator = PerEpochLoss(
                epoch_key=self._epoch_key, metric=self._metric, t_ignore=self._t_ignore,
            )
        return self._per_epoch_loss_tabulator

    @property
    def training(self) -> pd.Series:
        if not hasattr(self, "_training"):
            self._training = self._PER_EPOCH_LOSS_TABULATOR(
                self.metrics_df, stage=self._training_key
            )
        return self._training

    @property
    def validation(self) -> pd.Series:
        if not hasattr(self, "_validation"):
            self._validation = self._PER_EPOCH_LOSS_TABULATOR(
                self.metrics_df, stage=self._validation_key
            )
        return self._validation

    @property
    def best_validation(self):
        if not hasattr(self, "_best_validation"):
            self._best_validation = (
                self.validation.sort_values()
                .head(self._n_best)
                .to_frame()
                .rename({0: "validation_loss"}, axis=1)
            )
        return self._best_validation

    @property
    def _SAVED_CKPTS(self):
        if not hasattr(self, "_saved_ckpts"):
            self._saved_ckpts = list(self._version.ckpts.keys())
        return self._saved_ckpts

    @property
    def _BEST_SAVED_CKPTS(self):
        if not hasattr(self, "_best_saved_ckpts"):
            self._best_saved_ckpts = self.best_validation.loc[
                self.best_validation.index.isin(self._SAVED_CKPTS)
            ]
        return self._best_saved_ckpts

    @property
    def best_epoch(self):
        """best validation epoch"""
        if not hasattr(self, "_best_epoch"):
            self._best_epoch = self.validation.loc[
                self._BEST_SAVED_CKPTS.index
            ].idxmax()
        return self._best_epoch

    @property
    def best_ckpt(self) -> sdq.io.Checkpoint:
        """ """
        if not hasattr(self, "_best_ckpt"):
            self._best_ckpt = self._version.ckpts[self.best_epoch]
        return self._best_ckpt
    
    def __repr__(self) -> str:
        return f"""SummarizedLoss[{self._version._NAME}]"""
    


# -- import packages: ---------------------------------------------------------
import numpy as np
import pandas as pd


# -- function: ----------------------------------------------------------------
def rowwise_negative_cross_entropy(df1: pd.DataFrame, df2: pd.DataFrame):
    # Ensure that the dataframes have the same shape
    assert df1.shape == df2.shape, "DataFrames must have the same shape"

    df1 = df1.sort_index(axis=1)
    df2 = df2.sort_index(axis=1)

    # Convert the dataframes to numpy arrays for easier manipulation
    arr1 = df1.values
    arr2 = df2.values

    # Clip values to avoid log(0), which is undefined
    epsilon = 1e-12
    arr1 = np.clip(arr1, epsilon, 1.0 - epsilon)
    arr2 = np.clip(arr2, epsilon, 1.0 - epsilon)

    # Calculate negative cross-entropy row by row
    return -np.sum(arr2 * np.log(arr1), axis=1)




# -- import packages: ---------------------------------------------------------
import ABCParse
import matplotlib.pyplot as plt
import pandas as pd


# -- import local dependencies: -----------------------------------------------
from .. import parsers


# -- supporting data operational class: ---------------------------------------
class LossPlotData(ABCParse.ABCParse):
    def __init__(
        self, summarized_loss: parsers.SummarizedLoss, window: int = 50, *args, **kwargs
    ):
        self.__parse__(locals())

    def _smooth(self, values: pd.Series):
        return values.rolling(window=self._window, center=True).mean()

    @property
    def _TRAINING_LOSS(self):
        return self._summarized_loss.training

    @property
    def _VALIDATION_LOSS(self):
        return self._summarized_loss.validation

    @property
    def training(self):
        return {"x": self._TRAINING_LOSS.index, "y": self._TRAINING_LOSS.values}

    @property
    def validation(self):
        return {"x": self._VALIDATION_LOSS.index, "y": self._VALIDATION_LOSS.values}

    @property
    def smoothed_training(self):
        return self._smooth(self._TRAINING_LOSS).dropna()

    @property
    def smoothed_validation(self):
        return self._smooth(self._VALIDATION_LOSS).dropna()


# -- API-facing operational class: ------------------------------------------
class LossPlot(ABCParse.ABCParse):
    def __init__(
        self,
        summarized_loss,
        training_color: str = "#004e89",  # "#00a6ed"
        validation_color: str = "#fbb02d",  # "#ffb400",
        best_epoch_color: str = "#f6511d",
        window: int = 50,
        scatter_kwargs={
            "s": 5,
            "ec": "None",
            "alpha": 0.2,
            "zorder": 10,
            "rasterized": True,
        },
        plot_kwargs={"zorder": 20},
        ymin=500,
        ymax=2500,
    ):
        self.__parse__(locals())

        self.loss_pl_data = LossPlotData(
            summarized_loss=self._summarized_loss, window=self._window
        )

    @property
    def cmap(self):
        return {
            "training": self._training_color,
            "validation": self._validation_color,
            "best": self._best_epoch_color,
        }

    def _plot_raw(self, ax: plt.Axes, key: str):
        loss = getattr(self.loss_pl_data, key)
        ax.scatter(**loss, **self._scatter_kwargs, c=self.cmap[key])

    def _plot_smoothed(self, ax: plt.Axes, key: str):
        smooth_loss = getattr(self.loss_pl_data, f"smoothed_{key}")
        ax.plot(smooth_loss, **self._plot_kwargs, color=self.cmap[key])

    @property
    def best_epoch(self):
        return self._summarized_loss.best_epoch

    @property
    def _name(self):
        seed = str(self._summarized_loss._version).split("_")[-1]
        return f"Seed {seed}"

    def _plot_best_epoch(self, ax):

        ax.vlines(
            x=self.best_epoch,
            ymin=self._ymin,
            ymax=self._ymax,
            ls="--",
            color=self.cmap["best"],
            zorder=20,
        )
        ax.text(
            x=self.best_epoch - 50,
            y=self._ymax - 300,
            s=f"Best epoch: {self.best_epoch}",
            color=self.cmap["best"],
            ha="right",
            fontsize=8,
            zorder=21,
        )

    def _formatting(self, ax, lw: float = 0.5):

        ax.set_xlim(-200, 2700)
        ax.set_ylim(self._ymin, self._ymax)
        ax.grid(alpha=0.2, color="k", zorder=-10)
        [spine.set_lw(lw) for spine in list(ax.spines.values())]
        ax.xaxis.set_tick_params(width=lw)
        ax.yaxis.set_tick_params(width=lw)
        ax.set_title(self._name, fontsize=10)

    def forward(self, ax: plt.Axes, key: str):

        self._plot_raw(ax, key)
        self._plot_smoothed(ax, key)

    def __call__(self, ax: plt.Axes, lw: float = 0.5):
        """ """
        self.__update__(locals())

        self.ax = ax
        self.forward(ax, key="training")
        self.forward(ax, key="validation")
        self._plot_best_epoch(ax)
        self._formatting(ax, lw=lw)


# -- API-facing function: -----------------------------------------------------
def fit_loss(
    summarized_loss: parsers.SummarizedLoss,
    ax: plt.Axes,
    window: int = 50,
    lw: float = 0.5,
    *args,
    **kwargs,
):
    """ """

    loss_plot = LossPlot(
        summarized_loss=summarized_loss,
        window=window,
    )
    loss_plot(ax=ax, lw=lw)



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import autodevice
import lightning
import larry
import logging
import pathlib
import scdiffeq as sdq
import uuid
import wandb
import os

# -- import local dependencies: -----------------------------------------------
from . import funcs

# -- set type hints: ----------------------------------------------------------
from typing import Any, Dict, List, Literal, Optional, Union, Tuple

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- cls: ---------------------------------------------------------------------
class Runner(ABCParse.ABCParse):
    _script_name = pathlib.Path(__file__).name

    def __init__(
        self,
        h5ad_path: Union[pathlib.Path, str],
        project_name: str,
        wandb_api_key: str,
        time_key: str = "median_qbin",  # "t",  # "Time point",
        time_point: float = 2,
        weight_key: str = "W",
        ckpt_frequency=1,
        save_last_ckpt=True,
        keep_ckpts=3,
        monitor: str = "epoch_validation_loss",
    ) -> None:
        """
        Args:
            h5ad_path (Union[pathlib.Path, str]): Path to the h5ad file.
            project_name (str): Name of the project.
            time_key (str): Key for the time point.
            weight_key (str): Key for the weight.
            wandb_api_key (str): API key for wandb.
            ckpt_frequency (int): Frequency of saving checkpoints.
            save_last_ckpt (bool): Whether to save the last checkpoint.
            keep_ckpts (int): Number of checkpoints to keep.
            monitor (str): Metric to monitor.
        """
        self.__parse__(locals())

        wandb.login(key=self._wandb_api_key)

    @property
    def adata(self) -> anndata.AnnData:
        if not hasattr(self, "_ADATA"):
            self._ADATA = funcs.load_adata(
                h5ad_path=self._h5ad_path,
                weight_key=self._weight_key,
                time_key="Time point", # real time points
            )
        return self._ADATA

    @property
    def seeds(self) -> List[int]:
        if not hasattr(self, "_seeds_configured"):
            if hasattr(self, "_seeds") and (not self._seeds is None):
                self._seeds_configured = ABCParse.as_list(self._seeds)
            elif hasattr(self, "_n_seeds") and (not self._n_seeds is None):
                self._seeds_configured = range(self._n_seeds)
            else:
                self._seeds_configured = [0]
        return self._seeds_configured

    def _log_script_as_wandb_artifact(self, run_id: str) -> None:

        artifact = wandb.Artifact(
            name=f"{run_id}-{self._script_name}",
            type="run-script.py",
        )
        artifact.add_file(__file__)
        wandb.log_artifact(artifact)

    def _format_ckpt_fname(self, ckpt_fpath: pathlib.Path):
        return ckpt_fpath.name.split(".")[0].replace("=", "_").replace("-", ".")

    def _log_model_ckpt_as_wandb_artifact(self, seed, ckpt_fpath, run_id):

        ckpt_fname = self._format_ckpt_fname(ckpt_fpath)

        artifact = wandb.Artifact(
            name=f"model-ckpt-seed_{seed}-{ckpt_fname}-run_id-{run_id}", type="model"
        )
        artifact.add_file(ckpt_fpath)
        wandb.log_artifact(artifact)

        return ckpt_fname

    def _compose_model_params(self, seed: int) -> Dict[str,Any]:
        return {
            "adata": self.adata,
            "seed": seed,
            "latent_dim": self._latent_dim,
            "time_key": self._time_key,
            "use_key": self._use_key,
            "weight_key": self._weight_key,
            "mu_hidden": self._mu_hidden,
            "sigma_hidden": self._sigma_hidden,
            "mu_dropout": self._mu_dropout,
            "sigma_dropout": self._sigma_dropout,
            "DiffEq_type": self._diffeq_type,
            "batch_size": self._batch_size,
            "potential_type": self._potential_type,
            "coef_diffusion": self._coef_diffusion,
            "train_lr": self._train_lr,
            "train_step_size": self._train_step_size,
            "velocity_ratio_params": self._velocity_ratio_params,
            "monitor_hardware": True,
        }

    @property
    def _LOG_DIR(self):
        return self.model._metrics_path.parent

    @property
    def _CKPT_DIR(self):
        return self._LOG_DIR.joinpath("checkpoints/")

    @property
    def _CKPT_PATHS(self):
        return self._CKPT_DIR.glob("*.ckpt")

    def _update_wandb_params(
            self,
            run_id: str,
            params: Dict[str, Any],
            wandb_logger: wandb.sdk.wandb_run.Run,
        ) -> Tuple[wandb.sdk.wandb_run.Run, Dict[str, Any]]:
        """Update the wandb parameters and return the wandb logger and params."""
        # Create a copy of params excluding the adata object to prevent serialization issues
        wandb_params = {k: v for k, v in params.items() if k != "adata"}
        wandb_logger.experiment.config.update(wandb_params)
        wandb_logger.experiment.config.update({"h5ad_path": self._h5ad_path})
        wandb_logger.experiment.config.update({"run_id": run_id})
        params["logger"] = wandb_logger
        return wandb_logger, params

    def _setup_run(self, seed: int) -> Tuple[wandb.sdk.wandb_run.Run, dict, str]:
        run_id = str(uuid.uuid4()).split("-")[0]
        wandb_logger = lightning.pytorch.loggers.WandbLogger(project=self._project_name)
        self._log_script_as_wandb_artifact(run_id=run_id)
        params = self._compose_model_params(seed=seed)
        wandb_logger, params = self._update_wandb_params(run_id=run_id, params=params, wandb_logger=wandb_logger)
        logger.info(f"Model run [run_id: {run_id}, seed: {seed}]: CONSTRUCTED")
        return wandb_logger, params, run_id

    def _wandb_fate_prediction_metrics_logging(
        self, seed: int, run_id: str, ckpt_fpath
    ) -> None:
        logger.info(f"Model run [run_id: {run_id}, seed: {seed}]: SETUP WANDB FATE PREDICTION METRICS LOGGING")
        ckpt_fname = self._format_ckpt_fname(ckpt_fpath)
        artifact = wandb.Artifact(
            name=f"fate-prediction-metrics-seed_{seed}-{ckpt_fname}-run_id_{run_id}",
            type="model-prediction-task",
            metadata={"N": self._n_eval, "seed": seed, "run_id": run_id},
        )
        artifact.add_dir(self._FATE_PREDICTION_CALLBACK._CKPT_METRICS_PATH)
        wandb.log_artifact(artifact)

    def fit_model(self, seed: int, run_id: str, params: Dict[str, Any]):
        logger.info(f"Model run [run_id: {run_id}, seed: {seed}]: INITIALIZED")
        self.model = sdq.scDiffEq(**params)

        SWA_CALLBACK = sdq.callbacks.StochasticWeightAveraging(swa_lrs=self._swa_lrs)
        self._FATE_PREDICTION_CALLBACK = larry.callbacks.FatePredictionCallback(
            self.adata, N=self._n_eval
        )

        self.model.fit(
            train_epochs=self._train_epochs,
            ckpt_frequency=self._ckpt_frequency,
            save_last_ckpt=self._save_last_ckpt,
            keep_ckpts=self._keep_ckpts,
            monitor=self._monitor,
            mode="min",
            train_callbacks=[SWA_CALLBACK, self._FATE_PREDICTION_CALLBACK],
        )

        # -- note: callback automatically performs eval at pl_module.on_train_end()
        # Ensure checkpoint directory exists
        os.makedirs(self._CKPT_DIR, exist_ok=True)
        
        # Save the checkpoint manually
        ckpt_fname = f"on_train_end.epoch_{self.model.DiffEq.current_epoch}.ckpt"
        ckpt_fpath = self._CKPT_DIR.joinpath(ckpt_fname)
        self.model.DiffEq.trainer.save_checkpoint(ckpt_fpath)
        logger.info(f"Model run [run_id: {run_id}, seed: {seed}]: SAVED CHECKPOINT: {ckpt_fpath}")
        
        self._log_model_ckpt_as_wandb_artifact(seed=seed, ckpt_fpath=ckpt_fpath, run_id=run_id)
        self._wandb_fate_prediction_metrics_logging(
            seed=seed, run_id=run_id, ckpt_fpath=ckpt_fpath
        )

    def fate_prediction_evaluation(self, seed: int, ckpt_fpath: pathlib.Path, run_id: str) -> None:
        logger.info(f"Model run [run_id: {run_id}, seed: {seed}]: EVALUATING CKPT: {ckpt_fpath}")
        ckpt_fname = self._log_model_ckpt_as_wandb_artifact(
            seed=seed, ckpt_fpath=ckpt_fpath, run_id=run_id
        )

        self.model.DiffEq = sdq.io.load_diffeq(ckpt_fpath)
        self.model.to(autodevice.AutoDevice())
        self._FATE_PREDICTION_CALLBACK.forward(
            pl_module=self.model.DiffEq, ckpt_name=ckpt_fname, log_dir=self._LOG_DIR
        )
        self._wandb_fate_prediction_metrics_logging(seed=seed, run_id=run_id, ckpt_fpath=ckpt_fpath)

    def evaluate_model(self, seed: int, run_id: str) -> None:

        logger.info(f"Model run [run_id: {run_id}, seed: {seed}]: EVALUATING")

        for ckpt_fpath in self._CKPT_PATHS:
            self.fate_prediction_evaluation(
                seed=seed, ckpt_fpath=ckpt_fpath, run_id=run_id
            )
    def forward(self, seed: int) -> None:

        wandb.init(project=self._project_name)
        wandb_logger, params, run_id = self._setup_run(seed=seed)
        self.fit_model(seed=seed, run_id=run_id, params=params)
        self.evaluate_model(seed=seed, run_id=run_id)
        wandb.finish()

    def __call__(
        self,
        n_seeds: Optional[int] = None,
        seeds: Optional[Union[List[int], int]] = [],
        potential_type: Optional[Literal["fixed", "prior"]] = "fixed",
        train_epochs: int = 125,
        train_step_size: int = 25,
        mu_hidden: List[int] = [512, 512],
        sigma_hidden: List[int] = [32, 32],
        mu_dropout: float = 0,
        sigma_dropout: float = 0,
        train_lr=4e-5,
        n_eval: int = 2000,
        swa_lrs: float = 1e-8,
        batch_size: int = 512,
        coef_diffusion: float = 1,
        diffeq_type: str = "SDE",
        use_key: str = "X_pca",
        latent_dim: int = 50,
        velocity_ratio_params: Dict[str, Union[float, str]] = {
            "target": 1,
            "enforce": 0,
            "method": "square",
        },
    ) -> None:

        self.__update__(locals())

        for seed in self.seeds:
            self.forward(seed=seed)



# -- import packages: ---------------------------------------------------------
import anndata

# -- function: ----------------------------------------------------------------
def subset_lineage_adata(
        adata: anndata.AnnData,
        cell_idx: int,
        lineage_key: str = "clone_idx",
) -> anndata.AnnData:
    """Subset an AnnData object to a single lineage, based on a given cell index.
    
    Args:
        adata: AnnData object.
        cell_idx: Index of the cell to subset.
        lineage_key: Key in adata.obs that contains the lineage information.
        
    Returns:
        AnnData object containing the lineage.
    """
    clone_idx = adata.obs.loc[str(cell_idx)][lineage_key]
    adata_lin = adata[adata.obs[lineage_key] == clone_idx].copy()
    return adata_lin
